/*
 * CS 106X, Nick Troccoli
 * Based on code from Marty Stepp
 *
 * This program contains code for various sorting algorithms
 * including bogo sort, selection sort, insertion sort,
 * recursive merge sort, and quick sort.
 */

#include <fstream>
#include <iomanip>
#include <iostream>
#include "console.h"
#include "filelib.h"
#include "random.h"
#include "vector.h"
#include "timer.h"
#include "support.h"
#include "shuffle.h"

using namespace std;

// constants
static const int INITIAL_SIZE = 10;         // starting size of array
static const int MAX_SIZE     = 10000000;     // largest length of data to sort  (try 100k for selection, 2-3m for merge)

// function prototype declarations
void bogoSort(Vector<int>& v);
void selectionSort(Vector<int>& v);
void insertionSort(Vector<int>& v);
void mergeSort(Vector<int>& v);
void quickSort(Vector<int>& v);
void swap(Vector<int>& v, int i, int j);

int main() {
    for (int N = INITIAL_SIZE; N <= MAX_SIZE; N *= 2) {
        // cout << "Creating the vector..." << endl;
        Vector<int> v;
        fillRandomIntVector(v, N);

        // perform the sort time how long it takes
        Timer tim(true);
        quickSort(v);
        long elapsedTime = tim.stop();
        
        if (!isSorted(v)) {
            throw "Vector is not sorted!";
        }

        cout << "N=" << setw(8) << N << ", time=" << setw(5) << elapsedTime << endl;
    }

    return 0;
}

/*
 * Rearranges the elements of v into sorted order using
 * the bogo sort algorithm.
 */
void bogoSort(Vector<int>& v) {
    while (!isSorted(v)) {
        shuffle(v);         // from shuffle.h
    }
}

/*
 * Rearranges the elements of v into sorted order using
 * the selection sort algorithm.
 */
void selectionSort(Vector<int>& v) {
    for (int i = 0; i < v.size(); i++) {
        // walk across the array looking for the smallest value
        int smallestIndex = i;
        for (int j = i+1; j < v.size(); j++) {
            if (v[j] < v[smallestIndex]) {
                smallestIndex = j;
            }
        }
        // swap v[i] with v[smallestIndex]
        swap(v, i, smallestIndex);
    }
}

/*
 * Rearranges the elements of v into sorted order using
 * the insertion sort algorithm.
 */
void insertionSort(Vector<int>& v) {
    for (int i = 1; i < v.size(); i++) {
        int temp = v[i];

        // slide elements right to make room for v[i]
        int j = i;
        while (j >= 1 && v[j - 1] > temp) {
            v[j] = v[j - 1];
            j--;
        }
        v[j] = temp;
    }
}

/*
 * Rearranges the elements of v into sorted order using
 * a recursive merge sort algorithm.
 */
void mergeSort(Vector<int>& v) {
    if (v.size() >= 2) {
        // split array in half
        Vector<int> left  = v.subList(0, v.size() / 2);
        Vector<int> right = v.subList(v.size() / 2, v.size() - left.size());

        // sort halves
        mergeSort(left);
        mergeSort(right);

        // merge halves
        int i1 = 0;
        int i2 = 0;
        for (int i = 0; i < v.size(); i++) {
            if (i2 >= right.size() ||
                    (i1 < left.size() && left[i1] < right[i2])) {
                v[i] = left[i1++];
            } else {
                v[i] = right[i2++];
            }
        }

    } // else base case do nothing
}






















// Partitions a with elements < pivot on left and
// elements > pivot on right;
// returns index of element that should be swapped with pivot
int partition(Vector<int>& v, int i, int j, int pivot) {
    while (i <= j) {
        // move index markers i,j toward center
        // until we find a pair of out-of-order elements
        while (i <= j && v[i] < pivot) { i++; }
        while (i <= j && v[j] > pivot) { j--; }

        if (i <= j) {
            swap(v, i++, j--);
        }
    }
    return i;
}

void quickSortHelper(Vector<int>& v, int min, int max) {
    if (min >= max) {  // base case; no need to sort
        return;
    }

    // choose pivot; we'll use the first element (might be bad!)
    int pivot = v[min];
    swap(v, min, max);      // move pivot to end

    // partition the two sides of the array
    int middle = partition(v, min, max - 1, pivot);
    
    swap(v, middle, max);   // restore pivot to proper location

    // recursively sort the left and right partitions
    quickSortHelper(v, min, middle - 1);
    quickSortHelper(v, middle + 1, max);
}

/*
 * Rearranges the elements of v into sorted order using
 * a recursive quick sort algorithm.
 */
void quickSort(Vector<int>& v) {
    quickSortHelper(v, 0, v.size() - 1);
}



/*
runtimes:
bogo:
N=       4, time=    1
N=       5, time=    0
N=       6, time=    0
N=       7, time=    1
N=       8, time=   16
N=       9, time=   82
N=      10, time= 4051
N=      11, time= 7643

selection:
N=      10, time=    0
N=      20, time=    1
N=      40, time=    0
N=      80, time=    0
N=     160, time=    3
N=     320, time=    5
N=     640, time=   20
N=    1280, time=   76
N=    2560, time=  324
N=    5120, time=  857
N=   10240, time= 3455
N=   20480, time=13774

insertion:
N=      10, time=    0
N=      20, time=    0
N=      40, time=    0
N=      80, time=    0
N=     160, time=    1
N=     320, time=    4
N=     640, time=   15
N=    1280, time=   55
N=    2560, time=  212
N=    5120, time=  657
N=   10240, time= 2459
N=   20480, time=10275
N=   40960, time=39091

merge:
N=      10, time=    0
N=      20, time=    0
N=      40, time=    0
N=      80, time=    1
N=     160, time=    0
N=     320, time=    1
N=     640, time=    2
N=    1280, time=    5
N=    2560, time=   11
N=    5120, time=   20
N=   10240, time=   44
N=   20480, time=   99
N=   40960, time=  188
N=   81920, time=  329
N=  163840, time=  654
N=  327680, time= 1530
N=  655360, time= 3048
N= 1310720, time= 6259
N= 2621440, time=13095
N= 5242880, time=26848

quick:
N=      10, time=    0
N=      20, time=    0
N=      40, time=    0
N=      80, time=    0
N=     160, time=    0
N=     320, time=    1
N=     640, time=    0
N=    1280, time=    1
N=    2560, time=    3
N=    5120, time=    8
N=   10240, time=   15
N=   20480, time=   33
N=   40960, time=   72
N=   81920, time=  156
N=  163840, time=  242
N=  327680, time=  491
N=  655360, time= 1051
N= 1310720, time= 2252
N= 2621440, time= 4743
N= 5242880, time= 9639
*/
